#!/usr/bin/python

"""This file is part of asyncoro; see http://asyncoro.sourceforge.net
for details.

This program can be used to start discoro server processes so discoro
scheduler (see 'discoro.py') can send computations to these processes
for executing distributed communicating proceses (coroutines). All
coroutines in a process execute in the same thread in a process, so
multiple CPUs are not used by one process. If CPU intensive
computations are to be run on systems with multiple processors, then
this program should be run with multiple instances (see below for '-c'
option to this program).

See 'discoro_client.py' for an example.
"""

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2014 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

import os
import sys
import traceback
import time
import shutil

from asyncoro.discoro import MinPulseInterval, MaxPulseInterval, Scheduler
import asyncoro.disasyncoro as asyncoro
from asyncoro import Coro

__all__ = ['discoro_proc']


def discoro_proc(_discoro_name='discoro_proc'):
    # coroutine
    """Server process receives computations and runs coroutines for it.
    """

    _discoro_coro = asyncoro.AsynCoro.cur_coro()
    _discoro_coro.register(_discoro_name)
    asyncoro.AsynCoro.instance().dest_path = os.path.join(
        'discoro', 'proc%s' % (_discoro_config['name'][_discoro_config['name'].rindex('-'):]))
    _discoro_dest_path = asyncoro.AsynCoro.instance().dest_path
    if os.path.isdir(_discoro_dest_path):
        shutil.rmtree(_discoro_dest_path)
    os.makedirs(_discoro_dest_path)
    os.chdir(_discoro_dest_path)
    asyncoro.logger.debug('discoro "%s" started at %s; computation files will be saved in "%s"' %
                          (_discoro_config['name'], _discoro_coro.location, _discoro_dest_path))
    _discoro_req = _discoro_client = _discoro_auth = _discoro_msg = None
    _discoro_timer_coro = _discoro_pulse_coro = _discoro_timer_proc = _discoro_peer_status = None
    _discoro_monitor_coro = _discoro_monitor_proc = None
    _discoro_computation = _discoro_func = _discoro_var = None
    _discoro_job_coros = set()
    _discoro_globals = {}
    _discoro_locals = {}
    _discoro_globals.update(globals())
    _discoro_locals.update(locals())

    def _discoro_timer_proc(coro=None):
        coro.set_daemon()
        last_pulse = time.time()
        interval = None
        while True:
            reset = yield coro.sleep(interval)
            if reset:
                if not isinstance(_discoro_pulse_coro, Coro):
                    interval = None
                    continue
                interval = reset
                last_pulse = time.time()
                continue
            msg = {'ncoros': len(_discoro_job_coros), 'location': coro.location}
            if _discoro_pulse_coro.send(msg) == 0:
                last_pulse = time.time()
            else:
                if (time.time() - last_pulse) > (5 * interval):
                    asyncoro.logger.warning('scheduler is not reachable; '
                                            'closing computation "%s"' % _discoro_auth)
                    _discoro_coro.send({'req': 'close', 'auth': _discoro_auth})

    def _discoro_peer_status(coro=None):
        coro.set_daemon()
        while True:
            status = yield coro.receive()
            if isinstance(status, asyncoro.PeerStatus):
                if _discoro_computation and \
                       _discoro_computation.scheduler.location == status.location:
                    asyncoro.logger.debug('scheduler at %s quit; closing computation %s' %
                                          (status.location, _discoro_computation._auth))
                    msg = {'req': 'close', 'auth': _discoro_computation._auth}
                    _discoro_coro.send(msg)

    def _discoro_monitor_proc(coro=None):
        coro.set_daemon()
        while True:
            msg = yield coro.receive()
            if isinstance(msg, asyncoro.MonitorException):
                asyncoro.logger.debug('job %s done' % msg.args[0])
                _discoro_job_coros.discard(msg.args[0])
            else:
                asyncoro.logger.warning('%s: invalid monitor message ignored' % coro.location)

    _discoro_timer_coro = Coro(_discoro_timer_proc)
    _discoro_monitor_coro = Coro(_discoro_monitor_proc)
    asyncoro.AsynCoro.instance().peer_status(Coro(_discoro_peer_status))

    while True:
        _discoro_msg = yield _discoro_coro.receive()
        try:
            _discoro_req = _discoro_msg['req']
        except:
            asyncoro.logger.debug(traceback.format_exc())
            _discoro_req = None

        if _discoro_req == 'run':
            _discoro_client = _discoro_msg.get('client', None)
            _discoro_auth = _discoro_msg.get('auth', None)
            _discoro_func = _discoro_msg.get('func', None)
            if not isinstance(_discoro_client, Coro) or not _discoro_computation or \
               _discoro_auth != _discoro_computation._auth:
                asyncoro.logger.warning('invalid run: %s' % (type(_discoro_func)))
                if isinstance(_discoro_client, Coro):
                    _discoro_client.send(None)
                continue
            try:
                _discoro_func = asyncoro.unserialize(_discoro_func)
                if _discoro_func.code:
                    exec(_discoro_func.code, globals())
                job_coro = Coro(globals()[_discoro_func.name],
                                *(_discoro_func.args), **(_discoro_func.kwargs))
            except:
                asyncoro.logger.debug('invalid computation to run')
                # _discoro_func = Scheduler._Function(_discoro_func.name, None,
                #                                     _discoro_func.args, _discoro_func.kwargs)
                job_coro = (sys.exc_type, _discoro_func, traceback.format_exc())
            else:
                asyncoro.logger.debug('job %s created' % job_coro)
                _discoro_job_coros.add(job_coro)
                job_coro.notify(_discoro_monitor_coro)
                _discoro_var = _discoro_msg.get('notify', None)
                if isinstance(_discoro_var, Coro):
                    job_coro.notify(_discoro_var)
            _discoro_client.send(job_coro)
            del job_coro
        elif _discoro_req == 'setup':
            _discoro_client = _discoro_msg.get('client', None)
            _discoro_pulse_coro = _discoro_msg.get('pulse_coro', None)
            if not isinstance(_discoro_client, Coro) or not isinstance(_discoro_pulse_coro, Coro):
                continue
            if _discoro_computation is not None:
                asyncoro.logger.debug('invalid "setup" - busy')
                _discoro_client.send(-1)
                continue
            os.chdir(_discoro_dest_path)
            try:
                _discoro_computation = _discoro_msg['computation']
                exec(_discoro_computation._code, globals())
            except:
                _discoro_computation = None
                asyncoro.logger.warning('invalid computation')
                asyncoro.logger.debug(traceback.format_exc())
                _discoro_client.send(-1)
                continue
            if isinstance(_discoro_computation.pulse_interval, int) and \
               MinPulseInterval <= _discoro_computation.pulse_interval <= MaxPulseInterval:
                _discoro_computation.pulse_interval = _discoro_computation.pulse_interval
            else:
                _discoro_computation.pulse_interval = MinPulseInterval
            _discoro_timer_coro.resume(_discoro_computation.pulse_interval)
            asyncoro.logger.debug('computation "%s" from "%s"' %
                                  (_discoro_computation._auth, _discoro_msg['client']))
            _discoro_client.send(0)
        elif _discoro_req == 'close':
            _discoro_auth = _discoro_msg.get('auth', None)
            if not _discoro_computation or _discoro_auth != _discoro_computation._auth:
                continue
            asyncoro.logger.debug('deleting computation "%s"' % _discoro_computation._auth)
            # TODO: is it better to quit this process and start another?
            for _discoro_var in _discoro_job_coros:
                _discoro_var.terminate()
            _discoro_job_coros = set()
            for _discoro_var in list(globals()):
                if _discoro_var not in _discoro_globals:
                    # asyncoro.logger.warning('Removing global variable "%s"' % _discoro_var)
                    globals().pop(_discoro_var, None)
            globals().update(_discoro_globals)

            os.chdir(asyncoro.AsynCoro.instance().dest_path)
            shutil.rmtree(_discoro_dest_path)
            os.makedirs(_discoro_dest_path)
            _discoro_computation = _discoro_client = _discoro_pulse_coro = None
            _discoro_timer_coro.resume(None)
            os.chdir(_discoro_dest_path)
        elif _discoro_req == 'quit':
            break
        else:
            asyncoro.logger.warning('invalid command "%s" ignored' % _discoro_req)
            _discoro_client = _discoro_msg.get('client', None)
            if not isinstance(_discoro_client, Coro):
                continue
            _discoro_client.send(-1)

    # wait until all computations are done; process only 'close'
    while _discoro_job_coros:
        _discoro_msg = yield _discoro_coro.receive()
        try:
            _discoro_req = _discoro_msg['req']
        except:
            if isinstance(_discoro_msg, asyncoro.MonitorException):
                asyncoro.logger.debug('job %s done' % _discoro_msg.args[0])
                _discoro_job_coros.discard(_discoro_msg.args[0])
                continue
            else:
                asyncoro.logger.debug(traceback.format_exc())
            _discoro_req = None
        if _discoro_req == 'close':
            _discoro_auth = _discoro_msg.get('auth', None)
            if not _discoro_computation or _discoro_auth != _discoro_computation._auth:
                continue
            asyncoro.logger.debug('deleting computation "%s"' % _discoro_computation._auth)
            for _discoro_var in globals():
                if _discoro_var not in _discoro_globals:
                    # asyncoro.logger.warning('Removing global variable "%s"' % _discoro_var)
                    globals().pop(_discoro_var, None)
            for _discoro_var in locals:
                if _discoro_var not in _discoro_locals:
                    locals().pop(_discoro_var, None)
            break
        else:
            asyncoro.logger.warning('invalid command "%s" ignored' % _discoro_req)
            _discoro_client = _discoro_msg.get('client', None)
            if not isinstance(_discoro_client, Coro):
                continue
            _discoro_client.send(-1)
    os.chdir(asyncoro.AsynCoro.instance().dest_path)
    # shutil.rmtree(_discoro_dest_path, ignore_errors=True)


def _discoro_process(_discoro_config, _discoro_queue):
    _discoro_scheduler = asyncoro.AsynCoro(**_discoro_config)
    _discoro_coro = Coro(discoro_proc)
    _discoro_queue.put(asyncoro.serialize(_discoro_coro))
    _discoro_scheduler.finish()


if __name__ == '__main__':

    """
    If '-c' option is used with a positive number, discoro server is
    run that many instances, so CPU intesive coroutines can be invoked
    on them. If the number is negative, that many processors are not
    used (from the available processors). The default value for this
    option is '0', in which case all the available processors are
    used.

    '-n' option can be used to specify prefix name for asyncoro
    schedulers. This name is appended with hyphen followed by a unique
    number when AsynCoro is created. Note that the names in a cluster
    must be unique; otherwise, 'locate' may give inconsistent results.

    If '-d' option is used, debug logging is enabled.

    Remaining options are as per AsynCoro in disasyncoro module.
    """

    import logging
    import argparse
    import multiprocessing
    import socket

    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--cpus', dest='cpus', type=int, default=0,
                        help='number of CPUs/discoro instances to run; '
                        'if negative, that many CPUs are not used')
    parser.add_argument('-i', '--ip_addr', dest='node', default=None,
                        help='IP address or host name of this node')
    parser.add_argument('--ext_ip_addr', dest='ext_ip_addr', default=None,
                        help='External IP address to use (needed in case of NAT firewall/gateway)')
    parser.add_argument('-u', '--udp_port', dest='udp_port', type=int, default=51350,
                        help='UDP port number to use')
    parser.add_argument('-n', '--name', dest='name', default=None,
                        help='(symbolic) name given to AsynCoro schdulers on this node')
    parser.add_argument('--dest_path', dest='dest_path', default=None,
                        help='path where files sent by peers are stored')
    parser.add_argument('--max_file_size', dest='max_file_size', default=None, type=int,
                        help='maximum file size of any file transferred')
    parser.add_argument('-s', '--secret', dest='secret', default='',
                        help='authentication secret for handshake with peers')
    parser.add_argument('--certfile', dest='certfile', default=None,
                        help='file containing SSL certificate')
    parser.add_argument('--keyfile', dest='keyfile', default=None,
                        help='file containing SSL key')
    parser.add_argument('-d', '--debug', action='store_true', dest='loglevel', default=False,
                        help='if given, debug messages are printed')
    _discoro_config = vars(parser.parse_args(sys.argv[1:]))
    del parser

    if _discoro_config['loglevel']:
        asyncoro.logger.setLevel(logging.DEBUG)
    else:
        asyncoro.logger.setLevel(logging.INFO)
    del _discoro_config['loglevel']

    _discoro_cpus = multiprocessing.cpu_count()
    if _discoro_config['cpus'] > 0:
        if _discoro_config['cpus'] > _discoro_cpus:
            raise Exception('CPU count must be <= %s' % _discoro_cpus)
        _discoro_cpus = _discoro_config['cpus']
    elif _discoro_config['cpus'] < 0:
        if -_discoro_config['cpus'] >= _discoro_cpus:
            raise Exception('CPU count must be > -%s' % _discoro_cpus)
        _discoro_cpus += _discoro_config['cpus']
    del _discoro_config['cpus']

    _discoro_name = _discoro_config['name']
    if not _discoro_name:
        _discoro_name = socket.gethostname()
        if not _discoro_name:
            _discoro_name = 'discoro_proc'

    _discoro_queue = multiprocessing.Queue()
    _discoro_processes = []
    for _discoro_proc_id in range(2, _discoro_cpus + 1):
        _discoro_config['name'] = _discoro_name + '-%s' % _discoro_proc_id
        _discoro_processes.append(multiprocessing.Process(target=_discoro_process,
                                                          args=(_discoro_config, _discoro_queue)))
        _discoro_processes[-1].start()
        time.sleep(0.05)
    _discoro_proc_id = 1
    _discoro_config['name'] = _discoro_name + '-%s' % _discoro_proc_id
    _discoro_scheduler = asyncoro.AsynCoro(**_discoro_config)
    _discoro_coro = Coro(discoro_proc)

    while True:
        try:
            if sys.stdin.readline().strip().lower() in ('quit', 'exit'):
                break
        except KeyboardInterrupt:
            break

    asyncoro.logger.debug('terminating processes')
    # make sure all processes quit; otherwise, multiple keyboard
    # interrupts can leave processes hanging, with unpredictable behavior
    while True:
        try:
            while not _discoro_queue.empty():
                asyncoro.unserialize(_discoro_queue.get()).send({'req': 'quit'})
            _discoro_coro.send({'req': 'quit'})

            for _i, _discoro_proc in enumerate(_discoro_processes, start=1):
                if _discoro_proc.is_alive():
                    asyncoro.logger.info('  -- waiting for process %s to finish' % _i)
                    _discoro_proc.join()
            break
        except KeyboardInterrupt:
            print('waiting for processes to quit')
            continue



#!/usr/bin/env python

# Ping Pong Benchmark server used at http://nichol.as/asynchronous-servers-in-python

# test with httperf (adjust num-conns appropriately):
# httperf --timeout=1 --client=0/1 --server=localhost --port=8010 --uri=/ \
#     --rate=400 --send-buffer=4096 --recv-buffer=16384 --num-calls=1 --num-conns=2000

import sys, socket
import asyncoro

def process(conn, coro=None):
    msg = "HTTP/1.0 200 OK\r\nContent-Length: 5\r\n\r\nPong!\r\n"
    if sys.version_info.major >= 3:
        msg = bytes(msg, 'ascii')
    yield conn.sendall(msg)
    conn.close()

def server(host, port, coro=None):
    coro.set_daemon()
    sock = asyncoro.AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(5000)

    while True:
        conn, addr = yield sock.accept()
        asyncoro.Coro(process, conn)

if __name__ == '__main__':
    asyncoro.Coro(server, '127.0.0.1', 8010)

    while True:
        try:
            cmd = sys.stdin.readline().strip().lower()
            if not cmd or cmd == 'exit' or cmd == 'quit':
                break
        except KeyboardInterrupt:
            break



# server program for processing requests received with message passing
# (asynchronous concurrent programming) from remote client
# (tut_client.py) on same network;
# see http://asyncoro.sourceforge.net/tutorial.html for details.

import sys, random, logging
import asyncoro.disasyncoro as asyncoro

def server_proc(coro=None):
    coro.set_daemon()
    coro.register('server_coro')
    for x in range(10):
        msg = yield coro.receive()
        print('processing %s' % (msg))
    coro.unregister('server_coro')

asyncoro.logger.setLevel(logging.DEBUG)
scheduler = asyncoro.AsynCoro(udp_port=0)
server = asyncoro.Coro(server_proc)
while True:
    cmd = sys.stdin.readline().strip().lower()
    if cmd == 'quit' or cmd == 'exit':
        break



# This program uses status message notifications to submit jobs to
# discoro processes.

# StatusMessage must be imported in global scope as below; otherwise,
# unserializing status messages fails (if external scheduler is used)
from asyncoro.discoro import StatusMessage
import asyncoro.discoro as discoro
import asyncoro.disasyncoro as asyncoro

# this generator function is sent to remote server to run
# coroutines there
def rcoro_proc(n, coro=None):
    yield coro.sleep(n)
    raise StopIteration(n)

def client_proc(computation, njobs, coro=None):
    status = {'submitted': 0, 'done': 0}

    def submit_job(where, coro=None):
        arg = random.uniform(5, 20)
        rcoro = yield computation.run_at(where, rcoro_proc, arg)
        if isinstance(rcoro, asyncoro.Coro):
            print('%s processing %s' % (rcoro.location, arg))
        else:
            print('Job %s failed: %s' % (status['submitted'], str(rcoro)))
        status['submitted'] += 1

    computation.status_coro = coro
    # if scheduler is shared (i.e., running as program), nothing needs
    # to be done (its location can optionally be given to 'schedule');
    # othrwise, start scheduler:
    # discoro.Scheduler()
    if (yield computation.schedule()):
        raise Exception('Failed to schedule computation')
    # job submitter assumes that one process can run one coroutine at a time
    while True:
        msg = yield coro.receive()
        if isinstance(msg, asyncoro.MonitorException):
            # a process finished job
            rcoro = msg.args[0]
            if msg.args[1][0] == StopIteration:
                print('Remote coroutine %s finished with %s' % (rcoro.location, msg.args[1][1]))
            else:
                asyncoro.logger.warning('Remote coroutine %s terminated with "%s"' %
                                        (rcoro.location, str(msg.args[1])))
            status['done'] += 1
            if status['done'] == njobs:
                break
            if status['submitted'] < njobs:
                # schedule another job at this process
                asyncoro.Coro(submit_job, rcoro.location)
        elif isinstance(msg, StatusMessage):
            asyncoro.logger.debug('Node/Process status: %s, %s' % (msg.status, msg.location))
            if msg.status == discoro.Scheduler.ProcInitialized:
                # a new process is ready (if special initialization is
                # required for preparing process, schedule it)
                asyncoro.Coro(submit_job, msg.location)
        else:
            asyncoro.logger.debug('Ignoring status message %s' % str(msg))
    yield computation.close()

if __name__ == '__main__':
    import logging, random
    asyncoro.logger.setLevel(logging.DEBUG)
    computation = discoro.Computation([rcoro_proc])
    # run 10 jobs
    asyncoro.Coro(client_proc, computation, 10)



# This example shows how to use message passing, transferring files.
# The input files are assumed to be in current working directory as
# 'input0.dat', 'input1.dat', ..., 'input9.dat' (10 input files).
# 'submit_jobs_proc' creates jobs as and when a discoro process is found,
# or a process finishes a job. Each job is processed by a remote coroutine
# (rcoro) and a corresponding local coroutine (with 'client_proc')
# interacts with that rcoro to send the input data to where rcoro is,
# an instance of class C to give the information needed to process the
# file. Client then waits for rcoro to finish processing job, send the resulting
# output file in the form 'outputX.dat' and result (in this case, it is
# the return value of 'send_file').

# Note that the objective is to illustrate features, so implementation
# is not ideal. Error checking is skipped at a few places for brevity.

# StatusMessage must be imported in global scope as below; otherwise,
# unserializing status messages fails (if external scheduler is used)
from asyncoro.discoro import StatusMessage
import asyncoro.discoro as discoro
import asyncoro.disasyncoro as asyncoro

# objects of C are sent by a client to remote coroutine
class C(object):
    def __init__(self, i, n, client):
        self.i = i
        self.n = n
        self.client = client

# this generator function is used to create local coroutine (at the
# client) to communicate with a remote coroutine
def client_proc(job_id, rcoro, coro=None):
    # send input file; this will be saved to where discoro process is
    # (i.e., rcoro.location)
    if (yield asyncoro.AsynCoro().send_file(rcoro.location, 'input%s.dat' % job_id,
                                            timeout=30)) < 0:
        print('Could not send input data to %s' % rcoro.location)
        raise StopIteration(-1)
    # send info about input
    obj = C(job_id, random.uniform(5, 8), coro)
    if (yield rcoro.deliver(obj)) != 1:
        print('Could not send input to %s' % rcoro.location)
        raise StopIteration(-1)
    # rcoro sends result to this coroutine as message
    result = yield coro.receive()
    if result < 0:
        print('Processing %s failed' % obj.i)
        raise StopIteration(-1)
    # rcoro uses 'send_file'; this saves the file in asyncoro's
    # dest_path, not current working directory!
    output = os.path.join(asyncoro.AsynCoro().dest_path, 'output%s.dat' % obj.i)
    # if necessary, move file to os.getcwd()
    # os.rename(output, os.getcwd())
    print('%s processed' % (obj.i))

# this generator function is sent to remote server to run
# coroutines there
def rcoro_proc(coro=None):
    import os
    # receive object from client_proc coroutine
    cobj = yield coro.receive()
    # for given input file, create an output file with each line in
    # the output file computed as length of corresponding line in input file
    inp_file = 'input%s.dat' % cobj.i
    out_file = 'output%s.dat' % cobj.i
    with open(inp_file, 'r') as inp_fd:
        with open(out_file, 'w') as out_fd:
            for lineno, line in enumerate(inp_fd, start=1):
                out_fd.write('%d: %d\n' % (lineno, len(line)-1))
    # 'sleep' to simulate computing
    yield coro.sleep(cobj.n)
    # transfer the output file to client's asyncoro
    result = yield asyncoro.AsynCoro().send_file(cobj.client.location, out_file, overwrite=True,
                                                 timeout=30)
    cobj.client.send(result)
    os.remove(inp_file)
    os.remove(out_file)

def submit_jobs_proc(computation, njobs, coro=None):
    computation.status_coro = coro
    # if scheduler is shared (i.e., running as program), nothing needs
    # to be done (its location can optionally be given to 'schedule');
    # othrwise, start scheduler: discoro.Scheduler()
    discoro.Scheduler()
    if (yield computation.schedule()):
        raise Exception('Failed to schedule computation')
    submitted = 0
    done = 0
    while True:
        msg = yield coro.receive()
        if isinstance(msg, asyncoro.MonitorException):
            # a job is done
            rcoro = msg.args[0]
            done += 1
            if done == njobs:
                break
            if submitted < njobs:
                # submit another job
                rcoro = yield computation.run_at(rcoro.location, rcoro_proc)
                if isinstance(rcoro, asyncoro.Coro):
                    asyncoro.Coro(client_proc, submitted, rcoro)
                submitted += 1
        elif isinstance(msg, StatusMessage):
            # asyncoro.logger.debug('Node/Process status: %s, %s' % (msg.status, msg.location))
            if msg.status == discoro.Scheduler.ProcInitialized and submitted < njobs:
                # a new process found; submit a job
                rcoro = yield computation.run_at(msg.location, rcoro_proc)
                if isinstance(rcoro, asyncoro.Coro):
                    asyncoro.Coro(client_proc, submitted, rcoro)
                submitted += 1
        else:
            asyncoro.logger.debug('Ignoring status message %s' % str(msg))
    yield computation.close()

if __name__ == '__main__':
    import logging, random, os
    asyncoro.logger.setLevel(logging.DEBUG)
    # unlike in earlier examples, rcoro_proc is not sent with
    # computation; instead, it is sent each time a job is submitted,
    # which is a bit inefficient
    computation = discoro.Computation([C])
    # run 10 jobs
    asyncoro.Coro(submit_jobs_proc, computation, 10)



#!/usr/bin/env python

# run at least two instances of this program on either same node or
# multiple nodes on local network, along with 'chat_chan_server.py';
# text typed in a client is broadcast over a channel to all clients

import sys, logging
import asyncoro.disasyncoro as asyncoro

def recv_proc(client_id, coro=None):
    coro.set_daemon()
    while True:
        msg, who = yield coro.receive()
        if who == client_id:
            continue
        print('    %s %s' % (who, msg))

def send_proc(coro=None):
    # if server is in a remote network, use 'peer' as (optionally
    # enabling streaming for efficiency):
    # yield asyncoro.AsynCoro.instance().peer('server node/ip')
    server = yield asyncoro.Coro.locate('server')
    yield server.send(('join', coro))
    client_id = yield coro.receive()
    
    # channel is at same location as server coroutine
    channel = yield asyncoro.Channel.locate('channel', server.location)
    recv_coro = asyncoro.Coro(recv_proc, client_id)
    yield channel.subscribe(recv_coro)
    # since readline is synchronous (blocking) call, use async thread
    async_threads = asyncoro.AsyncThreadPool(1)
    while True:
        msg = yield async_threads.async_task(sys.stdin.readline)
        msg = msg.strip()
        if msg == 'quit':
            break
        channel.send((msg, client_id))
    server.send(('quit', client_id))
    yield channel.unsubscribe(recv_coro)

if __name__ == '__main__':
    # asyncoro.logger.setLevel(logging.DEBUG)
    asyncoro.Coro(send_proc)



#!/usr/bin/env python

# server program for client (tut_sock_client.py) sending requests with
# sockets (asynchronous network programming);
# see http://asyncoro.sourceforge.net/tutorial.html for details.

# run this program and then client either on same node. If they are on
# different computers, 'host' address must be changed appropriately.

import sys, socket
import asyncoro

def process(conn, coro=None):
    global n
    data = ''
    if sys.version_info.major >= 3:
        data = bytes(data, 'ascii')
    while True:
        data += yield conn.recv(128)
        if sys.version_info.major >= 3:
            if chr(data[-1]) == '/':
                break
        elif data[-1] == b'/':
            break
    conn.close()
    n += 1
    print('recieved: %s' % data)

def server(host, port, coro=None):
    coro.set_daemon()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock = asyncoro.AsyncSocket(sock)
    # sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(128)

    while True:
        conn, addr = yield sock.accept()
        asyncoro.Coro(process, conn)

n = 0
asyncoro.Coro(server, '127.0.0.1', 8010)

while True:
    cmd = sys.stdin.readline().strip().lower()
    if cmd == 'exit' or cmd == 'quit':
        break
print('n = %d' % n)



# Run 'discoronode.py' program to start processes to execute
# computations sent by this client, along with this program.

# Example where this client sends computation to remote discoro
# process to run as remote coroutines. Remote coroutines and client
# can use message passing to exchange data.

import sys, logging, random
import asyncoro.discoro as discoro
import asyncoro.disasyncoro as asyncoro

# objects of C are exchanged between client and servers
class C(object):
    def __init__(self, i):
        self.i = i
        self.n = None

    def __repr__(self):
        return '%d: %s' % (self.i, self.n)

# status messages indicating nodes, processes as well as remote
# coroutines finish status are sent to this coroutine ;see
# MonitorStatus for more comprehensive use cases
def status_proc(client, coro=None):
    procs_ready = 0
    coro.set_daemon()
    while True:
        msg = yield coro.receive()
        if isinstance(msg, discoro.StatusMessage):
            # print('Status: %s / %s' % (msg.location, msg.status))
            if msg.status == discoro.Scheduler.ProcInitialized:
                # wait until 2 processes ready
                procs_ready += 1
                if procs_ready == 2:
                    client.send('processes ready')
        elif isinstance(msg, asyncoro.MonitorException):
            if msg.args[1][0] != StopIteration:
                print('rcoro %s failed: %s / %s' % (msg.args[0], msg.args[1][0], msg.args[1][1]))

# this generator function is sent to remote discoro servers to run
# coroutines there
def compute(obj, client, coro=None):
    # obj is an instance of C
    import math
    # this coroutine and client can use message passing; client sends
    # data to this coro as a message
    n = yield coro.receive()
    print('process at %s received: %s' % (coro.location, n))
    obj.n = math.sqrt(n)
    yield coro.sleep(obj.n)
    # send result back to client
    yield client.deliver(obj, timeout=5)

def client_proc(computation, coro=None):
    # scheduler sends node / process status messages to status_coro
    computation.status_coro = asyncoro.Coro(status_proc, coro)

    # if no other clients use discoro, scheduler can be started in
    # client itself; alternately, scheduler can be run separately on a node
    discoro.Scheduler()

    # distribute computation to server
    if (yield computation.schedule()):
        raise Exception('schedule failed')

    # wait until processes are initialized; alternately, create
    # coroutines from 'status_proc' itself
    msg = yield coro.receive()
    assert msg == 'processes ready'

    rcoros = []
    # create remote coroutines
    for i in range(3):
        c = C(i)
        rcoro = yield computation.run(compute, c, coro)
        if isinstance(rcoro, asyncoro.Coro):
            rcoros.append(rcoro)
        else:
            print('failed to create remote coroutine for %s: %s' % (c, rcoro))

    # yield coro.sleep(10)
    # send data to remote coroutines (as messages)
    for rcoro in rcoros:
        rcoro.send(random.uniform(5, 20))

    # remote coroutines send replies as messages to this coro
    for rcoro in rcoros:
        reply = yield coro.receive()
        print('reply: %s' % (reply))

    yield computation.close()

if __name__ == '__main__':
    asyncoro.logger.setLevel(logging.DEBUG)
    # send generator function and class C (as the computation uses
    # objects of C)
    computation = discoro.Computation([compute, C])
    asyncoro.Coro(client_proc, computation)



#!/usr/bin/env python

# run at least two instances of this program on either same node or
# multiple nodes on local network, along with 'chat_sock_server.py';
# text typed in a client is sent to the all other clients

import sys, socket, logging, time
import asyncoro

def client_recv(conn, coro=None):
    conn = asyncoro.AsyncSocket(conn)
    while True:
        line = yield conn.recv_msg()
        if not line:
            break
        print(line.decode())

if __name__ == '__main__':
    asyncoro.logger.setLevel(logging.DEBUG)
    # host name or IP address of server is arg1
    if len(sys.argv) > 1:
        host = sys.argv[1]
    else:
        host = ''

    # port used by server is arg2
    if len(sys.argv) > 2:
        port = int(sys.argv[2])
    else:
        port = 1234

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    asyncoro.Coro(client_recv, sock)
    # wrap it with asyncoro's synchronous socket so 'send_msg' can be
    # used
    conn = asyncoro.AsyncSocket(sock, blocking=True)

    while True:
        line = sys.stdin.readline().strip()
        if line.lower() == 'exit' or line.lower() == 'quit':
            break
        conn.send_msg(line.encode())
    conn.shutdown(socket.SHUT_WR)



#!/usr/bin/env python

# client program for sending requests to server (tut_sock_server.py)
# with sockets (asynchronous network programming);
# see http://asyncoro.sourceforge.net/tutorial.html for details.

import sys, socket, random
import asyncoro

def client(host, port, n, coro=None):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock = asyncoro.AsyncSocket(sock)
    yield sock.connect((host, port))
    msg = '%d: ' % n + '-' * random.randint(100,300) + '/'
    if sys.version_info.major >= 3:
        msg = bytes(msg, 'ascii')
    yield sock.sendall(msg)
    sock.close()

# run 10 client coroutines
for n in range(10):
    asyncoro.Coro(client, '127.0.0.1', 8010, n)



#!/usr/bin/python

"""This file is part of asyncoro; see http://asyncoro.sourceforge.net
for details.

This module provides API for distributing generator functions and
their dependencies (files, Python functions, classes) to create remote
coroutines for distributed communicating processes. This should be
used with 'discoronode.py' program.

See 'discoro_client.py' for an example.
"""

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2014-2015 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

import os
import sys
import inspect
import hashlib
import collections
import time
import shutil
import atexit

import asyncoro.disasyncoro as asyncoro
from asyncoro import Coro, logger

__all__ = ['Scheduler', 'Computation']

MsgTimeout = 10
MinPulseInterval = MsgTimeout
MaxPulseInterval = 10 * MinPulseInterval

# status about nodes / processes are sent with this structure
StatusMessage = collections.namedtuple('StatusMessage', ['status', 'location'])

# for internal use only
_Function = collections.namedtuple('_Function', ['name', 'code', 'args', 'kwargs'])


class Scheduler(object):

    # status indications ('status' attribute of StatusMessage)
    NodeDiscovered = 1
    NodeInitialized = 2
    NodeClosed = 3
    NodeIgnore = 4
    NodeDisconnected = 5
    ProcDiscovered = 11
    ProcInitialized = 12
    ProcClosed = 13
    ProcIgnore = 14
    ProcDisconnected = 15

    """This class is for use by Computation class (see below) only.
    Other than the status messages above, none of its attributes are
    to be accessed directly.
    """

    __metaclass__ = asyncoro.MetaSingleton

    class _Node(object):

        def __init__(self, addr, scheduler):
            self.addr = addr
            self.procs = {}
            self.ncoros = 0
            self.status = None
            self.scheduler = scheduler

        def run(self, func, client):
            if self.status != Scheduler.NodeInitialized:
                client.send(None)
                raise StopIteration
            where = None
            load = None
            for proc in self.procs.itervalues():
                if proc.status != Scheduler.ProcInitialized:
                    continue
                if load is None or len(proc.coros) < load:
                    where = proc
                    load = len(proc.coros)
            if where:
                yield where.run(func, client)
            else:
                client.send(None)

    class _Proc(object):

        def __init__(self, name, location, scheduler):
            self.name = name
            self.location = location
            self.coros = {}
            self.done = {}
            self.status = None
            self.server = None
            self.xfer_files = []
            self.last_pulse = time.time()
            self.scheduler = scheduler

        def run(self, func, client):
            if self.status != Scheduler.ProcInitialized:
                raise StopIteration(None)
            node = self.scheduler._nodes.get(self.location.addr, None)
            computation = self.scheduler._cur_computation
            if not node or not computation:
                raise StopIteration(None)

            def _run(self, func, coro=None):
                self.server.send({'req': 'run', 'auth': computation._auth, 'func': func,
                                  'client': coro, 'notify': self.scheduler._status_coro})
                rcoro = yield coro.receive(timeout=computation.timeout)
                if isinstance(rcoro, Coro):
                    done = self.done.pop(str(rcoro), None)
                    if done is None:
                        # TODO: keep func too for fault-tolerance
                        self.coros[str(rcoro)] = rcoro
                        node.ncoros += 1
                    else:
                        rcoro = done
                raise StopIteration(rcoro)

            rcoro = yield Coro(_run, self, func).finish()
            yield client.deliver(rcoro)

    def __init__(self, **kwargs):
        self._nodes = {}
        self._cur_computation = None
        self.__cur_client_auth = None
        self.__pulse_interval = MinPulseInterval
        self.__sched_event = asyncoro.Event()
        self.__terminate = False

        kwargs['name'] = 'discoro_scheduler'
        self.asyncoro = asyncoro.AsynCoro.instance(**kwargs)
        if self.asyncoro.name == 'discoro_scheduler':
            self.__dest_path = os.path.join(self.asyncoro.dest_path, 'discoro', 'scheduler')
            if not os.path.isdir(self.__dest_path):
                os.makedirs(self.__dest_path)
            self.asyncoro.dest_path = self.__dest_path
        else:
            self.__dest_path = self.asyncoro.dest_path
        self.__scheduler_coro = Coro(self.__scheduler_proc)
        self.__client_coro = Coro(self.__client_proc)
        self.__timer_coro = Coro(self.__timer_proc)
        self._status_coro = Coro(self.__status_proc)
        self.asyncoro.peer_status(self._status_coro)
        atexit.register(self.__close)

    def __close(self):
        # TODO: wait for pending coroutines?
        self.__timer_coro.terminate()
        self.__client_coro.terminate()
        self._status_coro.terminate()

    def __status_proc(self, coro=None):
        coro.set_daemon()
        while True:
            msg = yield coro.receive()
            if isinstance(msg, asyncoro.MonitorException):
                rcoro = msg.args[0]
                node = self._nodes.get(rcoro.location.addr, None)
                if not node:
                    logger.warning('node %s is invalid' % rcoro.location.addr)
                    continue
                proc = node.procs.get(str(rcoro.location), None)
                if not proc:
                    logger.warning('proc "%s" is invalid' % (rcoro.location))
                    continue
                if proc.coros.pop(str(rcoro), None) is None:
                    logger.warning('rcoro "%s" is invalid at "%s"' % (rcoro, proc.location))
                    proc.done[str(rcoro)] = msg
                    continue
                if self._cur_computation and self._cur_computation.status_coro:
                    self._cur_computation.status_coro.send(msg)
                node.ncoros -= 1

            elif isinstance(msg, asyncoro.PeerStatus):
                computation = self._cur_computation
                if msg.status == asyncoro.PeerStatus.Online:
                    proc = Scheduler._Proc(msg.name, msg.location, self)
                    if computation and computation.status_coro:
                        status_msg = StatusMessage(Scheduler.ProcDiscovered, msg.location)
                        computation.status_coro.send(status_msg)
                    node = self._nodes.get(msg.location.addr, None)
                    if node:
                        node.procs[str(msg.location)] = proc
                        Coro(self.__setup_proc, proc, computation)
                    else:
                        node = Scheduler._Node(msg.location.addr, self)
                        self._nodes[msg.location.addr] = node
                        node.procs[str(msg.location)] = proc
                        if computation and computation.status_coro:
                            status_msg = StatusMessage(Scheduler.NodeDiscovered,
                                                       msg.location.addr)
                            computation.status_coro.send(status_msg)
                        Coro(self.__setup_node, node, computation)
                else:
                    # msg.status == asyncoro.PeerStatus.Offline
                    node = self._nodes.get(msg.location.addr, None)
                    if node:
                        proc = node.procs.pop(str(msg.location), None)
                        if proc:
                            if computation and computation.status_coro:
                                status_msg = StatusMessage(Scheduler.ProcDisconnected,
                                                           proc.location)
                                computation.status_coro.send(status_msg)
                            # TODO: (re)start process elsewhere (fault-tolerant)?
                            if node.procs:
                                Coro(self.__close_proc, proc)
                            else:
                                self._nodes.pop(proc.location.addr, None)
                                if computation and computation.status_coro:
                                    status_msg = StatusMessage(Scheduler.NodeDisconnected,
                                                               proc.location.addr)
                                    computation.status_coro.send(status_msg)
                                Coro(self.__close_node, node)
                    elif computation and msg.location == computation._pulse_coro.location:
                        logger.warning('client %s terminated; closing computation %s' %
                                       (msg.location, self.__cur_client_auth))
                        Coro(self.__close_computation)
            else:
                logger.warning('invalid status message ignored')

    def __timer_proc(self, coro=None):
        coro.set_daemon()
        client_pulse = time.time()
        proc_check = time.time()
        while True:
            msg = yield coro.receive(timeout=self.__pulse_interval)
            now = time.time()
            if isinstance(msg, dict):  # message from a node's proc
                ncoros = msg.get('ncoros', -1)
                if ncoros >= 0:
                    loc = msg.get('location', None)
                    if isinstance(loc, asyncoro.Location):
                        node = self._nodes.get(loc.addr, None)
                        if node:
                            proc = node.procs.get(str(loc), None)
                            if proc:
                                proc.last_pulse = now
                                if ncoros != len(proc.coros):
                                    logger.debug('process %s running %s coroutines, '
                                                 'scheduler running %s' %
                                                 (proc.location, ncoros, len(proc.coros)))
            if (now - client_pulse) > self.__pulse_interval and self._cur_computation:
                if self._cur_computation._pulse_coro.send('pulse') == 0:
                    client_pulse = now
                else:
                    if (now - client_pulse) > (5 * self.__pulse_interval):
                        logger.debug('client %s not responding; closing it' %
                                     self.__cur_client_auth)
                        Coro(self.__close_computation)

            if (now - proc_check) > (5 * self.__pulse_interval):
                for node in self._nodes.itervalues():
                    if node.status != Scheduler.NodeInitialized:
                        continue
                    for proc in node.procs.itervalues():
                        if proc.status != Scheduler.ProcInitialized:
                            continue
                        if (now - proc.last_pulse) > (5 * self.__pulse_interval):
                            logger.warning('Process %s is zombie!' % proc.location)
                            Coro(self.__close_proc, proc)

    def __run(self, func, client):
        host = None
        load = None
        for node in self._nodes.itervalues():
            if node.status != Scheduler.NodeInitialized:
                continue
            node_load = float(node.ncoros) / len(node.procs)
            if load is None or node_load < load:
                host = node
                load = node_load
        if host:
            yield host.run(func, client)
        else:
            client.send(None)

    @staticmethod
    def auth_code():
        # TODO: use uuid?
        return hashlib.sha1(os.urandom(10).encode('hex')).hexdigest()

    def __scheduler_proc(self, coro=None):
        coro.set_daemon()
        while not self.__terminate:
            if self._cur_computation:
                self.__sched_event.clear()
                yield self.__sched_event.wait()
                continue

            self._cur_computation, client = yield coro.receive()

            logger.debug('Computation %s scheduled' % self._cur_computation._auth)
            if isinstance(self._cur_computation.pulse_interval, int) and \
               MinPulseInterval <= self._cur_computation.pulse_interval <= MaxPulseInterval:
                self.__pulse_interval = self._cur_computation.pulse_interval
            else:
                self.__pulse_interval = MinPulseInterval

            self.__cur_client_auth = self._cur_computation._auth
            self._cur_computation._auth = Scheduler.auth_code()
            msg = {'resp': 'scheduled', 'auth': self.__cur_client_auth}
            if (yield client.deliver(msg, timeout=self._cur_computation.timeout)) != 1:
                logger.warning('client not reachable?')
                self._cur_client_auth = None
                self._cur_computation = None
                continue

            for node in self._nodes.itervalues():
                for i, proc in enumerate(node.procs.itervalues()):
                    if i == 0:
                        node.status = Scheduler.NodeDiscovered
                        if self._cur_computation.status_coro:
                            status_msg = StatusMessage(Scheduler.NodeDiscovered, proc.location.addr)
                            self._cur_computation.status_coro.send(status_msg)
                    proc.status = Scheduler.ProcDiscovered
                    if self._cur_computation.status_coro:
                        status_msg = StatusMessage(Scheduler.ProcDiscovered, proc.location)
                        self._cur_computation.status_coro.send(status_msg)
            for node in self._nodes.itervalues():
                # TODO: check if node is allowed
                if self._cur_computation:
                    Coro(self.__setup_node, node, self._cur_computation)

    def __client_proc(self, coro=None):
        coro.set_daemon()
        coro.register('discoro_scheduler')
        computations = {}
        while not self.__terminate:
            msg = yield coro.receive()
            req = msg.get('req', None)
            client = msg.get('client', None)
            auth = msg.get('auth', None)
            if not isinstance(client, Coro):
                logger.warning('Ignoring invalid client request "%s"' % req)
                continue

            if req == 'run':
                func = msg.get('func', None)
                if not func or self.__cur_client_auth != auth:
                    logger.warning('Ignoring invalid request to run computation')
                    client.send(None)
                    continue
                where = msg.get('where', None)
                if not where:
                    Coro(self.__run, func, client)
                elif isinstance(where, str):
                    node = self._nodes.get(where, None)
                    if node:
                        Coro(node.run, func, client)
                    else:
                        client.send(None)
                elif isinstance(where, asyncoro.Location):
                    node = self._nodes.get(where.addr)
                    if node:
                        proc = node.procs.get(str(where))
                        if proc:
                            Coro(proc.run, func, client)
                        else:
                            client.send(None)
                    else:
                        client.send(None)
                else:
                    client.send(None)

            elif req == 'run_each':
                where = msg.get('where', None)
                func = msg.get('func', None)
                if not func or self.__cur_client_auth != auth:
                    logger.warning('Ignoring invalid request to run computation')
                    where = None
                if where == 'node':
                    nodes = [node for node in self._nodes.itervalues()
                             if node.status == Scheduler.NodeInitialized]
                    if (yield client.deliver(len(nodes), self._cur_computation.timeout)) != 1:
                        continue
                    for node in nodes:
                        Coro(node.run, func, client)
                elif where == 'proc':
                    procs = [proc for node in self._nodes.itervalues()
                             if node.status == Scheduler.NodeInitialized
                             for proc in node.procs.itervalues()
                             if proc.status == Scheduler.ProcInitialized]
                    if (yield client.deliver(len(procs), self._cur_computation.timeout)) != 1:
                        continue
                    for proc in procs:
                        Coro(proc.run, func, client)
                else:
                    # node_procs
                    node = self._nodes.get(where)
                    if node and node.status == Scheduler.NodeInitialized:
                        procs = [proc for proc in node.procs.itervalues()
                                 if proc.status == Scheduler.ProcInitialized]
                    else:
                        procs = []
                    if (yield client.deliver(len(procs), self._cur_computation.timeout)) != 1:
                        continue
                    for proc in procs:
                        Coro(proc.run, func, client)

            elif req == 'schedule':
                try:
                    computation = asyncoro.unserialize(msg['computation'])
                    assert isinstance(computation, Computation) or \
                           computation.__class__.__name__ == 'Computation'
                    assert isinstance(computation._pulse_coro, Coro)
                    if computation._pulse_coro.location == self.asyncoro.location:
                        computation._pulse_coro._id = int(computation._pulse_coro._id)
                        if computation.status_coro:
                            computation.status_coro._id = int(computation.status_coro._id)
                except:
                    logger.warning('ignoring invalid computation request')
                    client.send(None)
                    continue
                while True:
                    computation._auth = Scheduler.auth_code()
                    if not os.path.exists(os.path.join(self.__dest_path, computation._auth)):
                        break
                try:
                    os.mkdir(os.path.join(self.__dest_path, computation._auth))
                except:
                    logger.debug('Could not create "%s"' %
                                 os.path.join(self.__dest_path, computation._auth))
                    client.send(None)
                    continue
                # TODO: save it on disk instead
                computations[computation._auth] = computation
                client.send(computation._auth)

            elif req == 'await':
                computation = computations.pop(auth, None)
                if not computation:
                    client.send(None)
                    continue
                if computation._pulse_coro.location.addr != self.asyncoro.location.addr:
                    computation._xfer_files = [os.path.join(self.__dest_path, computation._auth,
                                                            os.path.basename(xf))
                                               for xf in computation._xfer_files]
                for xf in computation._xfer_files:
                    if not os.path.isfile(xf):
                        logger.warning('File "%s" for computation %s is not valid' %
                                       (xf, computation._auth))
                        computation = None
                        break
                if computation is None:
                    client.send(None)
                else:
                    self.__scheduler_coro.send((computation, client))
                    self.__sched_event.set()

            elif req == 'close_computation':
                if self.__cur_client_auth != auth:
                    logger.warning('Ignoring invalid request to close computation')
                    continue
                Coro(self.__close_computation)

            elif req == 'nodes_list':
                # TODO: allowed to query anytime, even if current
                # computation is not the one querying?
                if self.__cur_client_auth == auth:
                    nodes = [node.addr for node in self._nodes.itervalues()
                             if node.status == Scheduler.NodeInitialized]
                else:
                    nodes = []
                client.send(nodes)

            elif req == 'procs_list':
                # TODO: allowed to query anytime, even if current
                # computation is not the one querying?
                if self.__cur_client_auth == auth:
                    procs = [proc.location for node in self._nodes.itervalues()
                             if node.status == Scheduler.NodeInitialized
                             for proc in node.procs.itervalues()
                             if proc.status == Scheduler.ProcInitialized]
                else:
                    procs = []
                client.send(procs)

            else:
                logger.warning('Ignoring invalid client request "%s"' % req)

    def __setup_node(self, node, computation, coro=None):
        if node.status == Scheduler.NodeIgnore:
            raise StopIteration(0)
        proc_setup_coros = []
        for proc in node.procs.itervalues():
            if proc.status in [Scheduler.ProcDiscovered, None]:
                proc_setup_coros.append(Coro(self.__setup_proc, proc, computation))
        for proc_setup_coro in proc_setup_coros:
            yield proc_setup_coro.finish()
        if node.status != Scheduler.NodeInitialized:
            for proc in node.procs.itervalues():
                if proc.status == Scheduler.ProcDiscovered:
                    node.status = Scheduler.NodeDiscovered
                    if computation and computation.status_coro:
                        computation.status_coro.send(StatusMessage(Scheduler.NodeDiscovered,
                                                                   proc.location.addr))
                elif proc.status == Scheduler.ProcInitialized:
                    node.status = Scheduler.NodeInitialized
                    if computation and computation.status_coro:
                        computation.status_coro.send(StatusMessage(Scheduler.NodeInitialized,
                                                                   proc.location.addr))
                    break
        raise StopIteration(0)

    def __setup_proc(self, proc, computation, coro=None):
        if proc.status == Scheduler.ProcIgnore:
            raise StopIteration(0)
        if proc.status == Scheduler.ProcInitialized:
            logger.warning('proc %s already initialized' % proc.location)
            raise StopIteration(0)
        proc.status = Scheduler.ProcIgnore
        if not proc.server:
            if computation:
                timeout = computation.timeout
            else:
                timeout = MsgTimeout
            proc.server = yield Coro.locate('discoro_proc', proc.location, timeout=timeout)
            if not isinstance(proc.server, Coro):
                logger.debug('server at %s is not valid' % (proc.location))
                # TODO: asuume temporary issue instead of removing it?
                node = self._nodes.get(proc.location.addr, None)
                if node:
                    node.procs.pop(str(proc.location), None)
                raise StopIteration(-1)
            if computation and computation.status_coro:
                computation.status_coro.send(StatusMessage(Scheduler.ProcDiscovered, proc.location))
        if not computation:
            proc.status = Scheduler.ProcDiscovered
            raise StopIteration(0)
        proc.server.send({'req': 'setup', 'client': coro, 'computation': computation,
                          'pulse_coro': self.__timer_coro})
        ret = yield coro.receive(timeout=computation.timeout)
        if ret:
            logger.warning('setup of %s failed: %s' % (proc.server, ret))
            raise StopIteration(ret)
        if computation._pulse_coro.location.addr != proc.location.addr:
            for xf in computation._xfer_files:
                reply = yield self.asyncoro.send_file(proc.location, xf,
                                                      timeout=computation.timeout)
                if reply < 0:
                    logger.debug('failed to transfer file %s: %s' % (xf, reply))
                    Coro(self.__close_proc, proc)
                    raise StopIteration(-1)
        proc.status = Scheduler.ProcInitialized
        proc.last_pulse = time.time()
        if computation.status_coro:
            computation.status_coro.send(StatusMessage(Scheduler.ProcInitialized, proc.location))
        raise StopIteration(0)

    def __close_node(self, node, coro=None):
        computation = self._cur_computation
        if not computation:
            logger.warning('Closing node %s ignored' % node.addr)
            raise StopIteration(-1)
        for proc in node.procs.values():
            yield self.__close_proc(proc, coro=coro)
        node.ncoros = 0
        node.status = Scheduler.NodeClosed
        if computation and computation.status_coro:
            computation.status_coro.send(StatusMessage(Scheduler.NodeClosed, node.addr))

    def __close_proc(self, proc, coro=None):
        computation = self._cur_computation
        if not computation or proc.status != Scheduler.ProcInitialized:
            logger.warning('Closing proc %s ignored' % proc.location)
            raise StopIteration(-1)
        if proc.coros:
            logger.warning('%s coros running at %s' % (len(proc.coros), proc.location))
        # TODO: check/indicate error?
        yield proc.server.deliver({'req': 'close', 'auth': computation._auth},
                                  timeout=computation.timeout)
        proc.status = Scheduler.ProcClosed
        proc.xfer_files = []
        proc.coros = {}
        proc.done = {}
        if computation and computation.status_coro:
            computation.status_coro.send(StatusMessage(Scheduler.ProcClosed, proc.location))
        raise StopIteration(0)

    def __close_computation(self, coro=None):
        for node in self._nodes.values():
            yield self.__close_node(node)
        if self.__cur_client_auth:
            computation_path = os.path.join(self.__dest_path, self.__cur_client_auth)
            if os.path.isdir(computation_path):
                shutil.rmtree(computation_path, ignore_errors=True)
        self._cur_computation = None
        self.__cur_client_auth = None
        self.__sched_event.set()
        raise StopIteration(0)


class Computation(object):
    """Packages components to distribute to remote asyncoro schedulers
    to create (remote) coroutines.
    """

    def __init__(self, components, status_coro=None, timeout=MsgTimeout,
                 pulse_interval=MinPulseInterval):
        """'components' is a list, each element of which should be
        either a module, a function, path name of a file, a class or
        an object (in which case the code for its class is sent).
        """
        if pulse_interval < MinPulseInterval or pulse_interval > MaxPulseInterval:
            raise Exception('"pulse_interval" must be at least %s and at most %s' %
                            (MinPulseInterval, MaxPulseInterval))
        if timeout < 1 or timeout > 120:
            raise Exception('"timeout" must be at least 1 and at most 120')
        if status_coro is not None and not isinstance(status_coro, Coro):
            raise Exception('status_coro must be coroutine')

        # TODO: add option to leave transferred files on the nodes
        # after computation is done (such as "cleanup" in dispy)?

        if not isinstance(components, list):
            components = [components]

        self._code = ''
        self._xfer_funcs = set()
        self._xfer_files = []
        self._node_xfers = []
        self.status_coro = status_coro
        self._auth = None
        self.scheduler = None
        self._pulse_coro = None
        self.pulse_interval = pulse_interval
        self.timeout = timeout
        depends = set()
        for dep in components:
            if isinstance(dep, str) or inspect.ismodule(dep):
                if inspect.ismodule(dep):
                    dep = dep.__file__
                    if dep.endswith('.pyc'):
                        dep = dep[:-1]
                    dep = os.path.abspath(dep)
                    if not (dep.endswith('.py') and os.path.isfile(dep)):
                        raise Exception('Invalid module "%s" - must be python source.' % dep)
                if dep in depends:
                    continue
                try:
                    fd = open(dep, 'rb')
                    fd.close()
                except:
                    raise Exception('File "%s" is not valid' % dep)
                depends.add(dep)
                self._xfer_files.append(dep)
            elif inspect.isgeneratorfunction(dep) or inspect.isfunction(dep) or \
               inspect.isclass(dep) or hasattr(dep, '__class__'):
                if inspect.isgeneratorfunction(dep) or inspect.isfunction(dep):
                    name = dep.func_name
                elif inspect.isclass(dep):
                    name = dep.__name__
                elif hasattr(dep, '__class__') and inspect.isclass(dep.__class__):
                    dep = dep.__class__
                    name = dep.__name__
                if name in depends:
                    continue
                depends.add(name)
                self._xfer_funcs.add(name)
                self._code += '\n' + inspect.getsource(dep).lstrip()
            else:
                raise Exception('Invalid computation: %s' % dep)
        # check code can be compiled
        compile(self._code, '<string>', 'exec')

    def schedule(self, location=None, timeout=None):
        """Schedule computation for execution. Must be used with
        'yield' as 'result = yield compute.schedule()'. If scheduler
        is executing other computations, this will block until
        scheduler processes them (computations are processed in the
        order submitted).
        """

        if self._auth is not None:
            raise StopIteration(0)

        if not self.scheduler:
            self.scheduler = yield Coro.locate('discoro_scheduler', location=location,
                                               timeout=self.timeout)
            if not isinstance(self.scheduler, asyncoro.Coro):
                raise StopIteration(-1)

        def _schedule(self, coro=None):
            self._pulse_coro = Coro(self._pulse_proc)
            msg = {'req': 'schedule', 'computation': asyncoro.serialize(self), 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) != 1:
                logger.debug('schedule failed')
                yield self.close()
                raise StopIteration(None)
            self._auth = yield coro.receive(timeout=self.timeout)
            if not isinstance(self._auth, str):
                yield self.close()
                raise StopIteration(-1)
            if isinstance(self._auth, str):
                atexit.register(self.close)
                if coro.location != self.scheduler.location:
                    for xf in self._xfer_files:
                        if (yield asyncoro.AsynCoro.send_file(
                           self.scheduler.location, xf, dir=self._auth, timeout=self.timeout)) < 0:
                            logger.warning('Could not send file "%s" to scheduler' % xf)
                            yield self.close()
                            raise StopIteration(-1)
            msg = {'req': 'await', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) != 1:
                yield self.close()
                raise StopIteration(-1)
            while True:
                resp = yield coro.receive(timeout=timeout)
                if isinstance(resp, dict) and resp.get('auth') == self._auth and \
                   resp.get('resp') == 'scheduled':
                    raise StopIteration(0)
                else:
                    logger.warning('invalid message ignored - waiting for scheduled')

        yield Coro(_schedule, self).finish()

    # TODO: add a way to send 'depends' to run methods?

    def run_at(self, where, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at 'where'. Must be used with 'yield' as 'rcoro =
        yield compute.run_at(loc, genf, ...)'. If the request is
        successful, 'rcoro' will be a (remote) coroutine.

        If 'where' is a string, it is assumed to be IP address of a
        node, in which case the function is scheduled at that node on
        a process with least load (i.e., process with least number of
        pending coroutines). If 'where' is a Location instance, it is
        assumed to be process location in which case the function is
        scheduled at that process.

        'func' must be generator function, as it is used to run
        coroutine at remote location.
        """
        if isinstance(func, str):
            name = func
        else:
            name = func.func_name

        if name in self._xfer_funcs:
            code = None
        else:
            # if not inspect.isgeneratorfunction(func):
            #     logger.warning('"%s" is not a valid generator function' % name)
            #     raise StopIteration(None)
            code = inspect.getsource(func).lstrip()

        def _run(self, coro=None):
            msg = {'req': 'run', 'auth': self._auth, 'where': where, 'client': coro,
                   'func': asyncoro.serialize(_Function(name, code, args, kwargs))}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                rcoro = yield coro.receive(self.timeout)
            else:
                rcoro = None
            raise StopIteration(rcoro)

        yield Coro(_run, self).finish()

    def run_each(self, where, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at each node or process. Must be used with
        'yield' as 'rcoros = yield compute.run_each(loc, genf,
        ...)'. 'rcoros' will be list of (remote) coroutines.

        'where' is same as in the case of 'run_at'; if it is string,
        the function is scheduled at every node and if it is a
        Location instance, the function is scheduled at every process
        (on every node).
        """
        if isinstance(func, str):
            name = func
        else:
            name = func.func_name

        if name in self._xfer_funcs:
            code = None
        else:
            # if not inspect.isgeneratorfunction(func):
            #     logger.warning('"%s" is not a valid generator function' % name)
            #     raise StopIteration([])
            code = inspect.getsource(func).lstrip()

        def _run(self, coro=None):
            msg = {'req': 'run_each', 'auth': self._auth, 'where': where, 'client': coro,
                   'func': asyncoro.serialize(_Function(name, code, args, kwargs))}
            n = yield self.scheduler.deliver(msg, timeout=self.timeout)
            if n != 1:
                raise StopIteration([])
            n = yield coro.receive(timeout=self.timeout)
            if not n:
                raise StopIteration([])
            rcoros = [(yield coro.receive(timeout=self.timeout)) for i in range(n)]
            # TODO: if failed, rcoro would be either None or tuple;
            # pass to user without filtering?
            rcoros = [rcoro for rcoro in rcoros if isinstance(rcoro, Coro)]
            raise StopIteration(rcoros)

        yield Coro(_run, self).finish()

    def run(self, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at a process with least load at a node with least
        load. Must be used with 'yield' as 'rcoro = yield
        compute.run(genf, ...)'. If the request is successful, 'rcoro'
        will be a (remote) coroutine.
        """
        yield self.run_at(None, func, *args, **kwargs)

    def run_nodes(self, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at a process with least load at every node. Must
        be used with 'yield' as 'rcoros = yield
        compute.run_nodes(genf, ...)'. 'rcoros' will be a list of
        (remote) coroutines.
        """
        yield self.run_each('node', func, *args, **kwargs)

    def run_procs(self, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at every process (at every node). Must be used
        with 'yield' as 'rcoros = yield compute.run_procs(genf,
        ...)'. 'rcoros' will be a list of (remote) coroutines.
        """
        yield self.run_each('proc', func, *args, **kwargs)

    def run_node_procs(self, addr, func, *args, **kwargs):
        """Run given generator function 'func' with arguments 'args'
        and 'kwargs' at every process at given node at 'addr'. Must be
        used with 'yield' as 'rcoros = yield
        compute.run_node_procs(addr, genf, ...)'. 'rcoros' will be a
        list of (remote) coroutines at that node.
        """
        yield self.run_each(addr, func, *args, **kwargs)

    # TODO: add 'map' methods to run with arguments as iterators
    # (e.g., list of tuples)

    def nodes(self):
        """Get list of addresses of nodes initialized for this
        computation. Must be used with 'yield' as 'yield
        compute.nodes()'.
        """

        def _nodes_list(self, coro=None):
            msg = {'req': 'nodes_list', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                yield coro.receive(self.timeout)
            else:
                raise StopIteration([])

        yield Coro(_nodes_list, self).finish()

    def procs(self):
        """Get list of Location instances of processes initialized for
        this computation. Must be used with 'yield' as 'yield
        compute.procs()'.
        """

        def _procs_list(self, coro=None):
            msg = {'req': 'procs_list', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                yield coro.receive(self.timeout)
            else:
                raise StopIteration([])

        yield Coro(_procs_list, self).finish()

    def close(self):
        """Close computation. Must be used with 'yield' as 'yield
        compute.close()'.
        """

        def _close(self, coro=None):
            msg = {'req': 'close_computation', 'auth': self._auth, 'client': coro}
            yield self.scheduler.deliver(msg, timeout=self.timeout)
            self._auth = None

        if self._auth:
            yield Coro(_close, self).finish()
        if self._pulse_coro:
            self._pulse_coro.send(None)

    def _pulse_proc(self, coro=None):
        """For internal use only.
        """
        last_pulse = time.time()
        while True:
            msg = yield coro.receive(timeout=(2 * self.pulse_interval))
            if msg == 'pulse':
                last_pulse = time.time()
            elif msg is None:
                logger.debug('scheduler not reachable?')
                if (time.time() - last_pulse) > (5 * self.pulse_interval):
                    logger.warning('scheduler is zombie!')
                if self._auth:
                    yield self.close()
                break
            else:
                logger.debug('ignoring invalid pulse message')


if __name__ == '__main__':
    """The scheduler can be started either within a client program (if
    no other client programs use the nodes simultaneously), or can be
    run on a node with the options described below (usually no options
    are necessary, so the scheduler can be strated with just
    'discoro.py')
    """

    import logging
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--ip_addr', dest='node', default=None,
                        help='IP address or host name of this node')
    parser.add_argument('--ext_ip_addr', dest='ext_ip_addr', default=None,
                        help='External IP address to use (needed in case of NAT firewall/gateway)')
    parser.add_argument('-u', '--udp_port', dest='udp_port', type=int, default=51350,
                        help='UDP port number to use')
    parser.add_argument('-n', '--name', dest='name', default=None,
                        help='(symbolic) name given to schduler')
    parser.add_argument('--max_file_size', dest='max_file_size', default=None, type=int,
                        help='maximum file size of any file transferred')
    parser.add_argument('-s', '--secret', dest='secret', default='',
                        help='authentication secret for handshake with peers')
    parser.add_argument('--certfile', dest='certfile', default=None,
                        help='file containing SSL certificate')
    parser.add_argument('--keyfile', dest='keyfile', default=None,
                        help='file containing SSL key')
    parser.add_argument('-d', '--debug', action='store_true', dest='loglevel', default=False,
                        help='if given, debug messages are printed')
    config = vars(parser.parse_args(sys.argv[1:]))
    del parser

    if not config['name']:
        config['name'] = 'discoro_scheduler'

    if config['loglevel']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    del config['loglevel']

    asyncoro.AsynCoro(**config)
    scheduler = Scheduler()
    while True:
        try:
            if sys.stdin.readline().strip().lower() in ('quit', 'exit'):
                break
        except KeyboardInterrupt:
            break

    logger.debug('terminating scheduler')



#!/usr/bin/env python

# client sends messages to a remote coroutine
# use with its server 'remote_coro_server.py'

import sys, logging
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

def sender(coro=None):
    # if server is in remote network, add it; set 'stream_send' to
    # True for streaming messages to it
    # yield scheduler.peer('remote.peer.ip', stream_send=True)
    rcoro = yield asyncoro.Coro.locate('server_coro')
    print('server is at %s' % rcoro.location)
    for x in range(10):
        rcoro.send('message %s' % x)

asyncoro.logger.setLevel(logging.DEBUG)
# scheduler = asyncoro.AsynCoro(secret='key')
asyncoro.Coro(sender)



"""Simple example demonstrating categorizing messages to coroutines.
"""

import sys, random, logging
import asyncoro

def client_proc(server, coro=None):
    for i in range(12):
        asyncoro.logger.debug('sending "%s"' % i)
        server.send(i)
        yield coro.sleep(random.uniform(0, 2))
    server.send('done')

def server_proc(coro=None):
    def even(msg):
        if msg % 2 == 0:
            return 'even'
        return None

    yield coro.sleep(3)
    # add 'even' method to categorize even numbers
    categories = asyncoro.CategorizeMessages(coro)
    categories.add(even)
    while True:
        # if messages arrive (or already queued) with 'even' category
        # within 1 second, receive them
        msg = yield categories.receive('even', timeout=1)
        if msg is None:
            # otherwise receive any message (without waiting, i.e.,
            # already queued)
            msg = yield categories.receive(timeout=0)
            asyncoro.logger.debug('         message: %s' % msg)
        else:
            asyncoro.logger.debug('    even message: %s' % msg)
            if msg == 8:
                # remove 'even' category, so future messages will no
                # longer be in 'even' category; note that there may
                # already be messages in 'even' category, so process
                # them if necessary
                categories.remove(even)
        if msg == 'done':
            break

if __name__ == '__main__':
    asyncoro.logger.setLevel(logging.DEBUG)
    server = asyncoro.Coro(server_proc)
    client = asyncoro.Coro(client_proc, server)



# Demonstratiion of hot swap feature

# when 'client' command is given, the function of coroutine 'client'
# ('client_proc') is replaced with 'client_proc2' function in the file
# 'hotswap_funcs.py'; similarly, if 'server' command is given,
# server's function is replaced

import sys, logging, random
import asyncoro

def server_proc(coro=None):
    coro.set_daemon()
    # indicate that this function can be swapped
    coro.hot_swappable(True)
    while True:
        try:
            n = yield coro.receive()
            print('%s received: %s' % (coro, n))
        except asyncoro.HotSwapException as exc:
            func = exc.args[0]
            print('\nnew function: %s, %s\n' % (func.__name__, func.gi_code.co_argcount))
            # replace only if new function meets certain criteria
            if func.__name__.startswith('server_proc'):
                raise
            else:
                print('\n** ignoring hot swap function %s' % (func.__name__))

def client_proc(server, i=1, coro=None):
    coro.set_daemon()
    coro.hot_swappable(True)
    # client doesn't process HotswapException
    while True:
        print('%s: sending %s' % (coro, i))
        server.send(i)
        i += 1
        yield coro.sleep(random.uniform(1, 3))

def swap(func_name, file_name, coro, *args, **kwargs):
    try:
        exec(open(file_name).read())
        func = locals()[func_name]
        coro.hot_swap(func, *args, **kwargs)
    except:
        print('failed to load "%s" from "%s"' % (func_name, file_name))

if __name__ == '__main__':
    asyncoro.logger.setLevel(logging.DEBUG)
    server = asyncoro.Coro(server_proc)
    client = asyncoro.Coro(client_proc, server)
    while True:
        try:
            cmd = sys.stdin.readline().strip().lower()
            if cmd.startswith('client'):
                swap('client_proc2', 'hotswap_funcs.py', client, server)
            elif cmd.startswith('server'):
                swap('server_proc2', 'hotswap_funcs.py', server, random.choice(['log', 'sqrt']))
            elif cmd == 'quit' or cmd == 'exit':
                break
        except KeyboardInterrupt:
            break



# Asynchronous pipe example with "communicate" method that is similar
# to Popen's "communicate". Same example is used to show how custom
# write/read processes can be provided to feed / read from the
# asynchronous pipe

# argv[1] must be a text file

import sys, os, logging, traceback, subprocess, platform
import asyncoro
import asyncoro.asyncfile
    
def communicate(input, coro=None):
    if platform.system() == 'Windows':
        # asyncfile.Popen must be used instead of subprocess.Popen
        pipe = asyncoro.asyncfile.Popen([r'\cygwin64\bin\sha1sum.exe'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    else:
        pipe = subprocess.Popen(['sha1sum'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    # convert pipe to asynchronous version
    async_pipe = asyncoro.asyncfile.AsyncPipe(pipe)
    # 'communicate' takes either the data or file descriptor with data
    # (if file is too large to read in full) as input
    input = open(input)
    stdout, stderr = yield async_pipe.communicate(input)
    print('communicate sha1sum: %s' % stdout)

def custom_feeder(input, coro=None):
    def write_proc(fin, pipe, coro=None):
        while True:
            data = yield os.read(fin.fileno(), 8*1024)
            if not data:
                break
            n = yield pipe.write(data, full=True)
            assert n == len(data)
        fin.close()
        pipe.stdin.close()

    def read_proc(pipe, coro=None):
        # output from sha1sum is small, so read until EOF
        data = yield pipe.stdout.read()
        pipe.stdout.close()
        raise StopIteration(data)

    if platform.system() == 'Windows':
        # asyncfile.Popen must be used instead of subprocess.Popen
        pipe = asyncoro.asyncfile.Popen([r'\cygwin64\bin\sha1sum.exe'],
                                        stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    else:
        pipe = subprocess.Popen(['sha1sum'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    async_pipe = asyncoro.asyncfile.AsyncPipe(pipe)
    reader = asyncoro.Coro(read_proc, async_pipe)
    writer = asyncoro.Coro(write_proc, open(input), async_pipe)
    stdout = yield reader.finish()
    print('     feeder sha1sum: %s' % stdout)

# asyncoro.logger.setLevel(logging.DEBUG)

# simpler version using 'communicate'
coro = asyncoro.Coro(communicate, sys.argv[1])
coro.value() # wait for it to finish

# alternate version with custom read and write processes
asyncoro.Coro(custom_feeder, sys.argv[1])



#!/usr/bin/env python

# chat server; must be used with 'chat_chan_client.py'

import sys, logging
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

asyncoro.MaxConnectionErrors = 3

def server_proc(coro=None):
    # to illustrate 'transform' function of channel, messages are modified
    def txfm_msgs(name, msg_cid):
        msg, client_id = msg_cid
        # assert name == 'channel'
        # e.g., drop shoutings
        if msg.isupper():
            return None
        if msg == 'joined':
            msg += ' :-)'
        elif msg == 'bye':
            msg = 'left :-('
        else:
            msg = 'says: %s' % msg
        return (msg, client_id)

    channel = asyncoro.Channel('channel', transform=txfm_msgs)
    channel.register()
    coro.set_daemon()
    coro.register('server')
    client_id = 1
    while True:
        cmd, who = yield coro.receive()
        # join/quit messages can be sent by clients themselves, but
        # for illustration server sends them instead
        if cmd == 'join':
            channel.send(('joined', client_id))
            who.send(client_id)
            client_id += 1
        elif cmd == 'quit':
            channel.send(('bye', who))
        elif cmd == 'terminate':
            break
    channel.unregister()
    coro.unregister()

if __name__ == '__main__':
    # asyncoro.logger.setLevel(logging.DEBUG)
    server = asyncoro.Coro(server_proc)
    while True:
        try:
            cmd = sys.stdin.readline().strip().lower()
            if cmd == 'quit' or cmd == 'exit':
                break
        except KeyboardInterrupt:
            break
    server.send(('terminate', None))
    server.value() # wait for server to finish



#!/usr/bin/env python

# server program for client sending requests to execute coroutines

# run this program and then client either on same node or different
# node on local network. Server and client can also be run on two
# different networks but client must call 'scheduler.peer' method
# appropriately.

import sys, logging
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

def rci_1(a, b=1, coro=None):
    asyncoro.logger.debug('running %s/%s with %s, %s', coro.name, id(coro), a, b)
    msg = yield coro.receive()
    if b % 2 == 0:
        yield coro.sleep(b)
        asyncoro.logger.debug('%s/%s done', coro.name, id(coro))
        # (remote) monitor (if any) gets this exception (to be
        # interpreted as normal termination)
        raise StopIteration(msg)
    else:
        # (remote) monitor (if any) gets this exception, too
        raise Exception('invalid invocation: %s' % b)

asyncoro.logger.setLevel(logging.DEBUG)
# 'secret' is set so only peers that use same secret can communicate
scheduler = asyncoro.AsynCoro(name='server', secret='test')
# register rci_1 so remote clients can request execution
rci1 = asyncoro.RCI(rci_1)
rci1.register()

while True:
    try:
        sys.stdin.readline()
    except KeyboardInterrupt:
        break



# program for broadcasting messages over channel; see
# http://asyncoro.sourceforge.net/tutorial.html for details.

import sys, random
import asyncoro

def seqsum(coro=None):
    # compute sum of numbers received over channel
    result = 0
    while True:
        msg = yield coro.receive()
        if msg is None:
            break
        result += msg
    print('sum: %f' % result)

def seqprod(coro=None):
    # compute product of numbers received over channel
    result = 1
    while True:
        msg = yield coro.receive()
        if msg is None:
            break
        result *= msg
    print('prod: %f' % result)

def client_proc(coro=None):
    channel = asyncoro.Channel('sum_prod')
    sum_coro = asyncoro.Coro(seqsum)
    prod_coro = asyncoro.Coro(seqprod)
    yield channel.subscribe(sum_coro)
    yield channel.subscribe(prod_coro)
    for x in range(4):
        r = random.uniform(0.5, 3)
        channel.send(r)
        print('sent %f' % r)
    channel.send(None)
    yield channel.unsubscribe(sum_coro)
    yield channel.unsubscribe(prod_coro)

asyncoro.Coro(client_proc)



#!/usr/bin/env python

import asyncoro, socket, logging, sys, time

def client_recv(conn, sender, coro=None):
    while True:
        line = yield conn.recv_msg()
        if not line:
            sender.terminate()
            break
        print(line.decode())

def client_send(conn, coro=None):
    thread_pool = asyncoro.AsyncThreadPool(1)
    while True:
        line = yield thread_pool.async_task(sys.stdin.readline)
        if not line:
            break
        yield conn.send_msg(line[:-1].encode())

if __name__ == '__main__':
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(('localhost', 1234))
    conn = asyncoro.AsynCoroSocket(sock)
    sender = asyncoro.Coro(client_send, conn)
    recvr = asyncoro.Coro(client_recv, conn, sender)
    sender.value()
    recvr.terminate()



# use with 'hotswap.py'

# 'server_proc2' replaces 'server' coroutine's function
def server_proc2(fname, coro=None):
    import math
    func = getattr(math, fname)
    coro.set_daemon()
    coro.hot_swappable(True)
    while True:
        try:
            n = yield coro.receive()
            print('%s received: %s,  %s=%.2f' % (coro, n, fname, func(n)))
        except asyncoro.HotSwapException as exc:
            gen = exc.args[0]
            if gen.__name__.startswith('server_proc'):
                print('\nreplacing with function: %s, %s\n' % \
                      (gen.__name__, gen.gi_code.co_argcount))
                raise
            else:
                print('\n** ignoring hot swap function %s' % (gen.__name__))

# 'client_proc2' replaces 'client' coroutine's function
def client_proc2(server, i=42, coro=None):
    coro.set_daemon()
    coro.hot_swappable(True)
    while True:
        print('%s: sending %s' % (coro, i))
        server.send(i)
        i += 2
        yield coro.sleep(random.uniform(2, 6))




#!/usr/bin/env python

# program for creating coroutines (asynchronous concurrent
# programming); see http://asyncoro.sourceforge.net/tutorial.html for
# details.

import sys, random, time
import asyncoro

def coro_proc(n, coro=None):
    s = random.uniform(0.5, 3)
    print('%f: coroutine %d sleeping for %f seconds' % (time.time(), n, s))
    yield coro.sleep(s)
    print('%f: coroutine %d terminating' % (time.time(), n))

for i in range(10):
    asyncoro.Coro(coro_proc, i)



# Run 'discoronode.py' program to start processes to execute
# computations sent by this client, along with this program.

# Example where this client sends computation to remote discoro
# process to run as remote coroutines. Remote coroutines and client
# can use message passing to exchange data.

import sys, logging, random
import asyncoro.discoro as discoro
import asyncoro.disasyncoro as asyncoro

# objects of C are exchanged between client and servers
class C(object):
    def __init__(self, i):
        self.i = i
        self.n = None

    def __repr__(self):
        return '%d: %.3f' % (self.i, self.n)

# this generator function is sent to remote discoro servers to run
# coroutines there
def compute(obj, client, coro=None):
    # obj is an instance of C
    import math
    # this coroutine and client can use message passing; client sends
    # data to this coro as a message
    n = yield coro.receive()
    print('process at %s received: %s' % (coro.location, n))
    obj.n = math.sqrt(n)
    yield coro.sleep(obj.n)
    # send result back to client
    yield client.deliver(obj, timeout=5)

def client_proc(computation, coro=None):
    # if no other clients use discoro, scheduler can be started in
    # client itself; alternately, scheduler can be run as a program
    discoro.Scheduler()

    # wait a bit for scheduler to detect processes; use MonitorStatus
    # to be notified of node / process state instead
    yield coro.sleep(1)

    # distribute computation to server
    if (yield computation.schedule()):
        raise Exception('schedule failed')

    # wait a bit for scheduler to send computation to processes
    yield coro.sleep(2)
    rcoros = []
    # create remote coroutines
    for i in range(3):
        c = C(i)
        rcoro = yield computation.run(compute, c, coro)
        if isinstance(rcoro, asyncoro.Coro):
            rcoros.append(rcoro)

    # send data to remote coroutines (as messages)
    for rcoro in rcoros:
        rcoro.send(random.uniform(5, 20))

    # remote coroutines send replies as messages to this 'coro'
    for rcoro in rcoros:
        reply = yield coro.receive()
        print('reply: %s' % (reply))

    yield computation.close()

if __name__ == '__main__':
    asyncoro.logger.setLevel(logging.DEBUG)
    # send generator function and class C (as the function uses
    # objects of C); 'depends' can include files, functions, objets
    computation = discoro.Computation([compute, C], timeout=5)
    asyncoro.Coro(client_proc, computation)



#!/usr/bin/env python

# server program where client sends messages to this
# server using coroutine send/receive

# run this program and then client either on same node or different
# node on local network. Server and client can also be run on two
# different networks but client must call 'scheduler.peer' method
# appropriately.

import sys, logging
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

def receiver(coro=None):
    coro.set_daemon()
    coro.register('server_coro')
    while True:
        msg = yield coro.receive()
        print('Received %s' % msg)

asyncoro.logger.setLevel(logging.DEBUG)
# call with 'udp_port=0' to start network services
# scheduler = asyncoro.AsynCoro(secret='key')

asyncoro.Coro(receiver)
while True:
    try:
        x = sys.stdin.readline()
    except KeyboardInterrupt:
        break

